@@ -6,6 +6,7 @@
  * Copyright (C) 2001 Havoc Pennington
  * Copyright (C) 2003 Kim Woelders
  * Copyright (C) 2003 Red Hat, Inc.
+ * Copyright (C) 2003, 2005-2007 Vincent Untz
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -23,6 +24,11 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
+
+#include <math.h>
+#include <glib/gi18n-lib.h>
+
 #include "pager.h"
 #include "workspace.h"
 #include "window.h"
@@ -31,7 +37,6 @@
 #include "pager-accessible-factory.h"
 #include "workspace-accessible-factory.h"
 #include "private.h"
-#include <math.h>
 
 /**
  * SECTION:pager
@@ -86,6 +91,7 @@ struct _WnckPagerPrivate
 };
 
 G_DEFINE_TYPE (WnckPager, wnck_pager, GTK_TYPE_WIDGET);
+#define WNCK_PAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), WNCK_TYPE_PAGER, WnckPagerPrivate))
 
 enum
 {
@@ -149,11 +155,17 @@ static gboolean wnck_pager_button_release (GtkWidget        *widget,
                                            GdkEventButton   *event);
 static gboolean wnck_pager_focus         (GtkWidget        *widget,
                                           GtkDirectionType  direction);
+static gboolean wnck_pager_query_tooltip (GtkWidget  *widget,
+                                          gint        x,
+                                          gint        y,
+                                          gboolean    keyboard_tip,
+                                          GtkTooltip *tooltip);
 static void workspace_name_changed_callback (WnckWorkspace *workspace,
                                              gpointer       data);
 
 static gboolean wnck_pager_window_state_is_relevant (int state);
-static gint wnck_pager_window_get_workspace   (WnckWindow  *window);
+static gint wnck_pager_window_get_workspace   (WnckWindow  *window,
+                                               gboolean     is_state_relevant);
 static void wnck_pager_queue_draw_workspace   (WnckPager   *pager,
 					       gint	    i);
 static void wnck_pager_queue_draw_window (WnckPager	   *pager,
@@ -182,21 +194,40 @@ static AtkObject* wnck_pager_get_accessible (GtkWidget *widget);
 static void
 wnck_pager_init (WnckPager *pager)
 {  
+  int i;
   static const GtkTargetEntry targets[] = {
     { "application/x-wnck-window-id", 0, 0}
   };
 
-  pager->priv = g_new0 (WnckPagerPrivate, 1);
+  pager->priv = WNCK_PAGER_GET_PRIVATE (pager);
 
   pager->priv->n_rows = 1;
   pager->priv->display_mode = WNCK_PAGER_DISPLAY_CONTENT;
   pager->priv->show_all_workspaces = TRUE;
   pager->priv->shadow_type = GTK_SHADOW_NONE;
+
   pager->priv->orientation = GTK_ORIENTATION_HORIZONTAL;
   pager->priv->workspace_size = 48;
-  pager->priv->bg_cache = NULL;
-  pager->priv->layout_manager_token = WNCK_NO_MANAGER_TOKEN;
+
+  for (i = 0; i < N_SCREEN_CONNECTIONS; i++)
+    pager->priv->screen_connections[i] = 0;
+
   pager->priv->prelight = -1;
+  pager->priv->prelight_dnd = FALSE;
+
+  pager->priv->dragging = FALSE;
+  pager->priv->drag_start_x = 0;
+  pager->priv->drag_start_y = 0;
+  pager->priv->drag_window = NULL;
+
+  pager->priv->bg_cache = NULL;
+
+  pager->priv->layout_manager_token = WNCK_NO_MANAGER_TOKEN;
+
+  pager->priv->dnd_activate = 0;
+  pager->priv->dnd_time = 0;
+
+  g_object_set (pager, "has-tooltip", TRUE, NULL);
 
   gtk_drag_dest_set (GTK_WIDGET (pager), 0, targets, G_N_ELEMENTS (targets), GDK_ACTION_MOVE);
   GTK_WIDGET_SET_FLAGS (GTK_WIDGET (pager), GTK_CAN_FOCUS);
@@ -207,7 +238,9 @@ wnck_pager_class_init (WnckPagerClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
   GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
-  
+
+  g_type_class_add_private (klass, sizeof (WnckPagerPrivate));
+
   object_class->finalize = wnck_pager_finalize;
 
   widget_class->realize = wnck_pager_realize;
@@ -227,6 +260,7 @@ wnck_pager_class_init (WnckPagerClass *klass)
   widget_class->drag_data_received = wnck_pager_drag_data_received;
   widget_class->drag_data_get = wnck_pager_drag_data_get;
   widget_class->drag_end = wnck_pager_drag_end;
+  widget_class->query_tooltip = wnck_pager_query_tooltip;
 }
 
 static void
@@ -247,8 +281,6 @@ wnck_pager_finalize (GObject *object)
       g_source_remove (pager->priv->dnd_activate);
       pager->priv->dnd_activate = 0;
     }
-
-  g_free (pager->priv);
   
   G_OBJECT_CLASS (wnck_pager_parent_class)->finalize (object);
 }
@@ -336,9 +368,11 @@ wnck_pager_unrealize (GtkWidget *widget)
 
   wnck_pager_clear_drag (pager);
   pager->priv->prelight = -1;
+  pager->priv->prelight_dnd = FALSE;
 
   wnck_screen_release_workspace_layout (pager->priv->screen,
                                         pager->priv->layout_manager_token);
+  pager->priv->layout_manager_token = WNCK_NO_MANAGER_TOKEN;
   
   wnck_pager_disconnect_screen (pager);
   pager->priv->screen = NULL;
@@ -650,13 +684,14 @@ wnck_pager_window_state_is_relevant (int state)
 }
 
 static gint
-wnck_pager_window_get_workspace (WnckWindow *window)
+wnck_pager_window_get_workspace (WnckWindow *window,
+                                 gboolean    is_state_relevant)
 {
   gint state;
   WnckWorkspace *workspace;
 
   state = wnck_window_get_state (window);
-  if (!wnck_pager_window_state_is_relevant (state))
+  if (is_state_relevant && !wnck_pager_window_state_is_relevant (state))
     return -1;
   workspace = wnck_window_get_workspace (window);
   if (workspace == NULL && wnck_window_is_pinned (window))
@@ -681,7 +716,7 @@ get_windows_for_workspace_in_bottom_to_top (WnckScreen    *screen,
   for (tmp = windows; tmp != NULL; tmp = tmp->next)
     {
       WnckWindow *win = WNCK_WINDOW (tmp->data);
-      if (wnck_pager_window_get_workspace (win) == workspace_num)
+      if (wnck_pager_window_get_workspace (win, TRUE) == workspace_num)
 	result = g_list_prepend (result, win);
     }
 
@@ -830,6 +865,45 @@ draw_window (GdkDrawable        *drawable,
   cairo_destroy (cr);
 }            
 
+static WnckWindow *
+window_at_point (WnckPager     *pager,
+                 WnckWorkspace *space,
+                 GdkRectangle  *space_rect,
+                 int            x,
+                 int            y)
+{
+  WnckWindow *window;
+  GList *windows;
+  GList *tmp;
+
+  window = NULL;
+
+  windows = get_windows_for_workspace_in_bottom_to_top (pager->priv->screen,
+                                                        space);
+
+  /* clicks on top windows first */
+  windows = g_list_reverse (windows);
+
+  for (tmp = windows; tmp != NULL; tmp = tmp->next)
+    {
+      WnckWindow *win = WNCK_WINDOW (tmp->data);
+      GdkRectangle winrect;
+
+      get_window_rect (win, space_rect, &winrect);
+
+      if (POINT_IN_RECT (x, y, winrect))
+        {
+          /* wnck_window_activate (win); */
+          window = win;
+          break;
+        }
+    }
+
+  g_list_free (windows);
+
+  return window;
+}
+
 static int
 workspace_at_point (WnckPager *pager,
                     int        x,
@@ -942,14 +1016,16 @@ wnck_pager_draw_workspace (WnckPager    *pager,
   GList *windows;
   GList *tmp;
   gboolean is_current;
-  WnckWorkspace *active_space;
+  WnckWorkspace *space;
   GtkWidget *widget;
   GtkStateType state;
   
+  space = wnck_screen_get_workspace (pager->priv->screen, workspace);
+  if (!space)
+    return;
+
   widget = GTK_WIDGET (pager);
-  active_space = wnck_screen_get_active_workspace (pager->priv->screen);
-  is_current = active_space &&
-    workspace == wnck_workspace_get_number (active_space);
+  is_current = (space == wnck_screen_get_active_workspace (pager->priv->screen));
 
   if (is_current)
     state = GTK_STATE_SELECTED;
@@ -972,23 +1048,119 @@ wnck_pager_draw_workspace (WnckPager    *pager,
                        GDK_RGB_DITHER_MAX,
                        0, 0);
     }
-  else if (active_space)
+  else
     {
-      int vx, vy, vw, vh; /* viewport */
-      double width_ratio, height_ratio;
       cairo_t *cr;
-      
-      width_ratio = rect->width / (double) wnck_workspace_get_width (active_space);
-      height_ratio = rect->height / (double) wnck_workspace_get_height (active_space);
-      vx = rect->x + width_ratio * wnck_workspace_get_viewport_x (active_space);
-      vy = rect->y + height_ratio * wnck_workspace_get_viewport_y (active_space);
-      vw = width_ratio * wnck_screen_get_width (pager->priv->screen);
-      vh = height_ratio * wnck_screen_get_height (pager->priv->screen);
 
       cr = gdk_cairo_create (widget->window);
-      gdk_cairo_set_source_color (cr, &widget->style->dark[state]);
-      cairo_rectangle (cr, vx, vy, vw, vh);
-      cairo_fill (cr);
+
+      if (!wnck_workspace_is_virtual (space))
+        {
+          gdk_cairo_set_source_color (cr, &widget->style->dark[state]);
+          cairo_rectangle (cr, rect->x, rect->y, rect->width, rect->height);
+          cairo_fill (cr);
+        }
+      else
+        {
+          //FIXME prelight for dnd in the viewport?
+          int workspace_width, workspace_height;
+          int screen_width, screen_height;
+          double width_ratio, height_ratio;
+          double vx, vy, vw, vh; /* viewport */
+
+          workspace_width = wnck_workspace_get_width (space);
+          workspace_height = wnck_workspace_get_height (space);
+          screen_width = wnck_screen_get_width (pager->priv->screen);
+          screen_height = wnck_screen_get_height (pager->priv->screen);
+
+          if ((workspace_width % screen_width == 0) &&
+              (workspace_height % screen_height == 0))
+            {
+              int i, j;
+              int active_i, active_j;
+              int horiz_views;
+              int verti_views;
+
+              horiz_views = workspace_width / screen_width;
+              verti_views = workspace_height / screen_height;
+
+              /* do not forget thin lines to delimit "workspaces" */
+              width_ratio = (rect->width - (horiz_views - 1)) /
+                            (double) workspace_width;
+              height_ratio = (rect->height - (verti_views - 1)) /
+                             (double) workspace_height;
+
+              if (is_current)
+                {
+                  active_i =
+                    wnck_workspace_get_viewport_x (space) / screen_width;
+                  active_j =
+                    wnck_workspace_get_viewport_y (space) / screen_height;
+                }
+              else
+                {
+                  active_i = -1;
+                  active_j = -1;
+                }
+
+              for (i = 0; i < horiz_views; i++)
+                {
+                  /* "+ i" is for the thin lines */
+                  vx = rect->x + (width_ratio * screen_width) * i + i;
+
+                  if (i == horiz_views - 1)
+                    vw = rect->width + rect->x - vx;
+                  else
+                    vw = width_ratio * screen_width;
+
+                  vh = height_ratio * screen_height;
+
+                  for (j = 0; j < verti_views; j++)
+                    {
+                      /* "+ j" is for the thin lines */
+                      vy = rect->y + (height_ratio * screen_height) * j + j;
+
+                      if (j == verti_views - 1)
+                        vh = rect->height + rect->y - vy;
+
+                      if (active_i == i && active_j == j)
+                        gdk_cairo_set_source_color (cr,
+                                                    &widget->style->dark[GTK_STATE_SELECTED]);
+                      else
+                        gdk_cairo_set_source_color (cr,
+                                                    &widget->style->dark[GTK_STATE_NORMAL]);
+                      cairo_rectangle (cr, vx, vy, vw, vh);
+                      cairo_fill (cr);
+                    }
+                }
+            }
+          else
+            {
+              width_ratio = rect->width / (double) workspace_width;
+              height_ratio = rect->height / (double) workspace_height;
+
+              /* first draw non-active part of the viewport */
+              gdk_cairo_set_source_color (cr, &widget->style->dark[GTK_STATE_NORMAL]);
+              cairo_rectangle (cr, rect->x, rect->y, rect->width, rect->height);
+              cairo_fill (cr);
+
+              if (is_current)
+                {
+                  /* draw the active part of the viewport */
+                  vx = rect->x +
+                    width_ratio * wnck_workspace_get_viewport_x (space);
+                  vy = rect->y +
+                    height_ratio * wnck_workspace_get_viewport_y (space);
+                  vw = width_ratio * screen_width;
+                  vh = height_ratio * screen_height;
+
+                  gdk_cairo_set_source_color (cr, &widget->style->dark[GTK_STATE_SELECTED]);
+                  cairo_rectangle (cr, vx, vy, vw, vh);
+                  cairo_fill (cr);
+                }
+            }
+        }
+
       cairo_destroy (cr);
     }
 
@@ -1057,7 +1229,7 @@ wnck_pager_draw_workspace (WnckPager    *pager,
       cairo_set_line_width (cr, 1.0);
       cairo_rectangle (cr,
 		       rect->x + 0.5, rect->y + 0.5,
-		       rect->width - 1, rect->height - 1);
+		       MAX (0, rect->width - 1), MAX (0, rect->height - 1));
       cairo_stroke (cr);
       cairo_destroy (cr);
     }
@@ -1177,33 +1349,11 @@ wnck_pager_button_press (GtkWidget      *widget,
     }
 
   if (space && (pager->priv->display_mode != WNCK_PAGER_DISPLAY_NAME))
-  {
-    GList *windows;
-    GList *tmp;
-
-    windows = get_windows_for_workspace_in_bottom_to_top (pager->priv->screen,
-                                                          space);
- 
-    /* clicks on top windows first */
-    windows = g_list_reverse (windows);
-
-    for (tmp = windows; tmp != NULL; tmp = tmp->next)
-      {
-        WnckWindow *win = WNCK_WINDOW (tmp->data);
-        GdkRectangle winrect;
-
-        get_window_rect (win, &workspace_rect, &winrect);
-
-        if (POINT_IN_RECT (event->x, event->y, winrect))
-          {
-            /* wnck_window_activate (win); */
-            pager->priv->drag_window = win;
-            break;
-          }
-      }
-
-    g_list_free (windows);
-  }
+    {
+      pager->priv->drag_window = window_at_point (pager, space,
+                                                  &workspace_rect,
+                                                  event->x, event->y);
+    }
 
   return TRUE;
 }
@@ -1247,7 +1397,7 @@ wnck_pager_queue_draw_window (WnckPager  *pager,
 {
   gint workspace;
 
-  workspace = wnck_pager_window_get_workspace (window);
+  workspace = wnck_pager_window_get_workspace (window, TRUE);
   if (workspace == -1)
     return;
 
@@ -1443,8 +1593,8 @@ wnck_pager_drag_motion_leave (GtkWidget          *widget,
 static void
 wnck_drag_clean_up (WnckWindow     *window,
 		    GdkDragContext *context,
-		    gboolean	    clean_up_window,
-		    gboolean	    clean_up_context);
+		    gboolean	    clean_up_for_context_destroy,
+		    gboolean	    clean_up_for_window_destroy);
 
 static void
 wnck_drag_context_destroyed (gpointer  windowp,
@@ -1523,11 +1673,19 @@ wnck_drag_source_destroyed (gpointer  contextp,
 static void
 wnck_drag_clean_up (WnckWindow     *window,
 		    GdkDragContext *context,
-		    gboolean	    clean_up_window,
-		    gboolean	    clean_up_context)
+		    gboolean	    clean_up_for_context_destroy,
+		    gboolean	    clean_up_for_window_destroy)
 {
-  if (clean_up_window)
+  if (clean_up_for_context_destroy)
     {
+      GtkWidget *drag_source;
+
+      drag_source = g_object_get_data (G_OBJECT (context),
+                                       "wnck-drag-source-widget");
+      if (drag_source)
+        g_object_weak_unref (G_OBJECT (drag_source),
+                             wnck_drag_source_destroyed, context);
+
       g_object_weak_unref (G_OBJECT (window),
                            wnck_drag_window_destroyed, context);
       if (g_signal_handlers_disconnect_by_func (window,
@@ -1535,19 +1693,9 @@ wnck_drag_clean_up (WnckWindow     *window,
 	g_assert_not_reached ();
     }
 
-  if (clean_up_context)
+  if (clean_up_for_window_destroy)
     {
-      GtkWidget *drag_source;
-
-      drag_source = g_object_get_data (G_OBJECT (context),
-                                       "wnck-drag-source-widget");
-      if (drag_source)
-        {
-          g_object_steal_data (G_OBJECT (context), "wnck-drag-source-widget");
-          g_object_weak_unref (G_OBJECT (drag_source),
-                               wnck_drag_source_destroyed, context);
-        }
-
+      g_object_steal_data (G_OBJECT (context), "wnck-drag-source-widget");
       g_object_weak_unref (G_OBJECT (context),
                            wnck_drag_context_destroyed, window);
     }
@@ -1711,7 +1859,7 @@ wnck_pager_focus (GtkWidget        *widget,
  * Does nothing.
  *
  * Since: 2.2
- * Deprecated:
+ * Deprecated:2.20:
  */
 void
 wnck_pager_set_screen (WnckPager  *pager,
@@ -1719,6 +1867,59 @@ wnck_pager_set_screen (WnckPager  *pager,
 {
 }
 
+static gboolean
+wnck_pager_query_tooltip (GtkWidget  *widget,
+                          gint        x,
+                          gint        y,
+                          gboolean    keyboard_tip,
+                          GtkTooltip *tooltip)
+{
+  int i;
+  WnckPager *pager;
+  WnckScreen *screen;
+  WnckWorkspace *space;
+  char *name;
+
+  pager = WNCK_PAGER (widget);
+  screen = pager->priv->screen;
+
+  i = workspace_at_point (pager, x, y, NULL, NULL);
+  space = wnck_screen_get_workspace (screen, i);
+  if (!space)
+    return GTK_WIDGET_CLASS (wnck_pager_parent_class)->query_tooltip (widget,
+                                                                      x, y,
+                                                                      keyboard_tip,
+                                                                      tooltip);
+
+  if (wnck_screen_get_active_workspace (screen) == space)
+    {
+      WnckWindow *window;
+      GdkRectangle workspace_rect;
+
+      get_workspace_rect (pager, i, &workspace_rect);
+
+      window = window_at_point (pager, space, &workspace_rect, x, y);
+
+      if (window)
+        name = g_strdup_printf (_("Click to start dragging \"%s\""),
+                                wnck_window_get_name (window));
+      else
+        name = g_strdup_printf (_("Current workspace: \"%s\""),
+                                wnck_workspace_get_name (space));
+    }
+  else
+    {
+      name = g_strdup_printf (_("Click to switch to \"%s\""),
+                              wnck_workspace_get_name (space));
+    }
+
+  gtk_tooltip_set_text (tooltip, name);
+
+  g_free (name);
+
+  return TRUE;
+}
+
 /**
  * wnck_pager_new:
  * @screen: deprecated argument, can be %NULL.
@@ -1892,6 +2093,8 @@ wnck_pager_set_display_mode (WnckPager            *pager,
   if (pager->priv->display_mode == mode)
     return;
 
+  g_object_set (pager, "has-tooltip", mode != WNCK_PAGER_DISPLAY_NAME, NULL);
+
   pager->priv->display_mode = mode;
   gtk_widget_queue_resize (GTK_WIDGET (pager));
 }
@@ -2045,7 +2248,16 @@ window_state_changed_callback     (WnckWindow      *window,
                                    gpointer         data)
 {
   WnckPager *pager = WNCK_PAGER (data);
-  wnck_pager_queue_draw_window (pager, window);
+
+  /* if the changed state changes the visibility in the pager, we need to
+   * redraw the whole workspace. wnck_pager_queue_draw_window() might not be
+   * enough */
+  if (!wnck_pager_window_state_is_relevant (changed))
+    wnck_pager_queue_draw_workspace (pager,
+                                     wnck_pager_window_get_workspace (window,
+                                                                      FALSE));
+  else
+    wnck_pager_queue_draw_window (pager, window);
 }
 
 static void
